/////////////////////////////////////////////////////////////////////
////                                                             ////
////  FASU                                                       ////
////  Floating point Add/Subtract Unit (Single precision)        ////
////                                                             ////
////  Author: Rudolf Usselmann                                   ////
////          russelmann@hotmail.com                             ////
////                                                             ////
/////////////////////////////////////////////////////////////////////
////                                                             ////
//// Copyright (C) 2000 Rudolf Usselmann                         ////
////                    russelmann@hotmail.com                   ////
////                                                             ////
//// This source file may be used and distributed without        ////
//// restriction provided that this copyright statement is not   ////
//// removed from the file and that any derivative work contains ////
//// the original copyright notice and the associated disclaimer.////
////                                                             ////
//// THIS SOURCE FILE IS PROVIDED "AS IS" AND WITHOUT ANY        ////
//// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT           ////
//// LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND   ////
//// FITNESS FOR A PARTICULAR PURPOSE.                           ////
////                                                             ////
/////////////////////////////////////////////////////////////////////


`timescale 1ns / 10ps

module fasu(
	clk,
	add,
	opa,
	opb,
	out,
	inf,
	nan
	);
input		clk;
input		add;
input	[31:0]	opa, opb;
output	[31:0]	out;
output		inf, nan;

////////////////////////////////////////////////////////////////////////
//
// Local Wires
//
reg	[31:0]	opa_r, opb_r;		// Input operand registers
reg		add_r;
reg	[31:0]	out;			// Output

wire		signa, signb;		// alias to opX sign
wire	[7:0]	expa, expb;		// alias to opX exponenet
wire	[22:0]	fracta, fractb;		// alias to opX fraction
wire		expa_lt_expb;		// expa is larger than expb indicator
wire		opb_lt_opa;		// opb is larger than opa indicator
wire	[7:0]	exp_out_dn;		// exponent output denormalized
wire	[7:0]	exp_small, exp_large;
wire	[7:0]	exp_diff;		// Numeric difference of the two exponents
wire	[23:0]	adj_op, adj_op_out;	// Fraction adjustment: input, output
wire	[23:0]	opa_tmp, opb_tmp;	// Fraction selection after normalising
wire	[23:0]	opa_tmp2, opb_tmp2;	// Fraction Sorting out
reg	[23:0]	opa_tmp2_r, opb_tmp2_r;	// Fraction Sorting out register
reg		add_tmp;		// opration (add/sub)
reg		add_tmp_r;		// opration (add/sub) register
wire	[24:0]	fract_out_dn;		// fraction output (denormalized)
reg	[24:0]	fract_out_dn_r;		// fraction output (denormalized) register
reg		sign_out;		// sign output
reg		inf, nan;		// Result INF & NaN Indicators

////////////////////////////////////////////////////////////////////////
//
// Input Registers & Aliases
//

reg	[31:0]	opa_r0, opb_r0;
reg		add_r0;
wire		infa_a, infa_b;
wire		infb_a, infb_b;

always @(posedge clk)
	opa_r <= #1 opa;

always @(posedge clk)
	opb_r <= #1 opb;

always @(posedge clk)
	add_r <= #1 add;

assign  signa = opa_r[31];
assign  signb = opb_r[31];
assign   expa = opa_r[30:23];
assign   expb = opb_r[30:23];
assign fracta = opa_r[22:0];
assign fractb = opb_r[22:0];

assign  infa_a = (expa == 8'hff);
assign  infa_b = (fracta == 23'h0000);
assign  infb_a = (expb == 8'hff);
assign  infb_b = (fractb == 23'h0000);

////////////////////////////////////////////////////////////////////////
//
// Normalize operator
//

assign expa_lt_expb = expa > expb;		// expa is larger than expb

// ---------------------------------------------------------------------
// Calculate the difference between the smaller and larger exponen

assign exp_small  = expa_lt_expb ? expb : expa;
assign exp_large  = expa_lt_expb ? expa : expb;
assign exp_diff   = exp_large - exp_small;
assign exp_out_dn = exp_large;

// ---------------------------------------------------------------------
// Adjust the smaller fraction

assign adj_op[22:0] = expa_lt_expb ? fractb : fracta;
assign adj_op[23]   = 1'b1;
assign adj_op_out   = adj_op >> exp_diff;	// FIX ME ???

// ---------------------------------------------------------------------
// Select operands for add/sub

assign opa_tmp = expa_lt_expb ? {1'b1, fracta} : adj_op_out;
assign opb_tmp = expa_lt_expb ? adj_op_out : {1'b1, fractb};

// ---------------------------------------------------------------------
// Sort operands (for sub only)

assign opb_lt_opa = opb_tmp > opa_tmp;		// opb is larger than opa
assign opa_tmp2 = opb_lt_opa ? opb_tmp : opa_tmp;
assign opb_tmp2 = opb_lt_opa ? opa_tmp : opb_tmp;

// ---------------------------------------------------------------------
// Decode Add/Sub operation

// add: 1=Add; 0=Subtract
always @(signa or signb or add_r)
   case({signa, signb, add_r})		// synopsys full_case parallel_case
   
   	// Add
	3'b0_0_1: add_tmp = 1;
	3'b0_1_1: add_tmp = 0;
	3'b1_0_1: add_tmp = 0;
	3'b1_1_1: add_tmp = 1;
	
	// Sub
	3'b0_0_0: add_tmp = 0;
	3'b0_1_0: add_tmp = 1;
	3'b1_0_0: add_tmp = 1;
	3'b1_1_0: add_tmp = 0;
   endcase

// ---------------------------------------------------------------------
// Determine sign for the output

// sign: 0=Posetive Number; 1=Negative Number
always @(signa or signb or add_r or opb_lt_opa)
   case({signa, signb, add_r})		// synopsys full_case parallel_case

   	// Add
	3'b0_0_1: sign_out = 0;
	3'b0_1_1: sign_out = opb_lt_opa;
	3'b1_0_1: sign_out = !opb_lt_opa;
	3'b1_1_1: sign_out = 1;

	// Sub
	3'b0_0_0: sign_out = opb_lt_opa;
	3'b0_1_0: sign_out = 0;
	3'b1_0_0: sign_out = 1;
	3'b1_1_0: sign_out = !opb_lt_opa;
   endcase


////////////////////////////////////////////////////////////////////////
//
// Add/Sub
reg		sign_out_r1, sign_out_r2;
reg	[7:0]	exp_out_dn_r1, exp_out_dn_r2;
reg		infa_a_r, infa_b_r;
reg		infb_a_r, infb_b_r;
reg		inf_r;
reg		nan_r;


always @(posedge clk)
	sign_out_r1 <= #1 sign_out;

always @(posedge clk)
	opa_tmp2_r <= #1 opa_tmp2;

always @(posedge clk)
	opb_tmp2_r <= #1 opb_tmp2;

always @(posedge clk)
	add_tmp_r <= #1 add_tmp;

always @(posedge clk)
	exp_out_dn_r1 <= #1 exp_out_dn;

always @(posedge clk)
	infa_a_r <= #1 infa_a;

always @(posedge clk)
	infa_b_r <= #1 infa_b;

always @(posedge clk)
	infb_a_r <= #1 infb_a;

always @(posedge clk)
	infb_b_r <= #1 infb_b;

assign fract_out_dn = add_tmp_r ? (opa_tmp2_r + opb_tmp2_r) : (opa_tmp2_r - opb_tmp2_r);

always @(posedge clk)
	fract_out_dn_r <= #1 fract_out_dn;

always @(posedge clk)
	sign_out_r2 <= #1 sign_out_r1;

always @(posedge clk)
	exp_out_dn_r2 <= #1 exp_out_dn_r1;

always @(posedge clk)
	inf_r <= #1 (infa_a_r & infa_b_r) | (infb_a_r & infb_b_r);

always @(posedge clk)
	nan_r <= #1 (infa_a_r & !infa_b_r) | (infb_a_r & !infb_b_r);
	
////////////////////////////////////////////////////////////////////////
//
// Normalize Result
//
reg	[4:0]	ld_zeros;
reg	[5:0]	exp_adj;
wire	[24:0]	fract_out;
wire	[7:0]	exp_out;
wire	[31:0]	out_d;

// Count Number of zeros in fract_out_dn from msb to lsb
always @(fract_out_dn_r)
   casex(fract_out_dn_r)	// synopsys full_case parallel_case
	25'b1_????_????_????_????_????_????: ld_zeros = 0;
	
	25'b0_1???_????_????_????_????_????: ld_zeros = 1;
	25'b0_01??_????_????_????_????_????: ld_zeros = 2;
	25'b0_001?_????_????_????_????_????: ld_zeros = 3;
	25'b0_0001_????_????_????_????_????: ld_zeros = 4;
	
	25'b0_0000_1???_????_????_????_????: ld_zeros = 5;
	25'b0_0000_01??_????_????_????_????: ld_zeros = 6;
	25'b0_0000_001?_????_????_????_????: ld_zeros = 7;
	25'b0_0000_0001_????_????_????_????: ld_zeros = 8;
	
	25'b0_0000_0000_1???_????_????_????: ld_zeros = 9;
	25'b0_0000_0000_01??_????_????_????: ld_zeros = 10;
	25'b0_0000_0000_001?_????_????_????: ld_zeros = 11;
	25'b0_0000_0000_0001_????_????_????: ld_zeros = 12;

	25'b0_0000_0000_0000_1???_????_????: ld_zeros = 13;
	25'b0_0000_0000_0000_01??_????_????: ld_zeros = 14;
	25'b0_0000_0000_0000_001?_????_????: ld_zeros = 15;
	25'b0_0000_0000_0000_0001_????_????: ld_zeros = 16;

	25'b0_0000_0000_0000_0000_1???_????: ld_zeros = 17;
	25'b0_0000_0000_0000_0000_01??_????: ld_zeros = 18;
	25'b0_0000_0000_0000_0000_001?_????: ld_zeros = 19;
	25'b0_0000_0000_0000_0000_0001_????: ld_zeros = 20;

	25'b0_0000_0000_0000_0000_0000_1???: ld_zeros = 21;
	25'b0_0000_0000_0000_0000_0000_01??: ld_zeros = 22;
	25'b0_0000_0000_0000_0000_0000_001?: ld_zeros = 23;
	25'b0_0000_0000_0000_0000_0000_0001: ld_zeros = 24;
   endcase

assign fract_out = fract_out_dn_r << ld_zeros;
 
// Determine how much to add to exp_out
always @(fract_out_dn_r)
   casex(fract_out_dn_r)	// synopsys full_case parallel_case
	25'b1_????_????_????_????_????_????: exp_adj =  1;
	
	25'b0_1???_????_????_????_????_????: exp_adj =  0;
	25'b0_01??_????_????_????_????_????: exp_adj = -1;
	25'b0_001?_????_????_????_????_????: exp_adj = -2;
	25'b0_0001_????_????_????_????_????: exp_adj = -3;

	25'b0_0000_1???_????_????_????_????: exp_adj = -4;
	25'b0_0000_01??_????_????_????_????: exp_adj = -5;
	25'b0_0000_001?_????_????_????_????: exp_adj = -6;
	25'b0_0000_0001_????_????_????_????: exp_adj = -7;
	
	25'b0_0000_0000_1???_????_????_????: exp_adj = -8;
	25'b0_0000_0000_01??_????_????_????: exp_adj = -9;
	25'b0_0000_0000_001?_????_????_????: exp_adj = -10;
	25'b0_0000_0000_0001_????_????_????: exp_adj = -11;

	25'b0_0000_0000_0000_1???_????_????: exp_adj = -12;
	25'b0_0000_0000_0000_01??_????_????: exp_adj = -13;
	25'b0_0000_0000_0000_001?_????_????: exp_adj = -14;
	25'b0_0000_0000_0000_0001_????_????: exp_adj = -15;

	25'b0_0000_0000_0000_0000_1???_????: exp_adj = -16;
	25'b0_0000_0000_0000_0000_01??_????: exp_adj = -17;
	25'b0_0000_0000_0000_0000_001?_????: exp_adj = -18;
	25'b0_0000_0000_0000_0000_0001_????: exp_adj = -19;

	25'b0_0000_0000_0000_0000_0000_1???: exp_adj = -20;
	25'b0_0000_0000_0000_0000_0000_01??: exp_adj = -21;
	25'b0_0000_0000_0000_0000_0000_001?: exp_adj = -22;
	25'b0_0000_0000_0000_0000_0000_0001: exp_adj = -23;
   endcase

assign exp_out = exp_out_dn_r2 + {exp_adj[5], exp_adj[5], exp_adj};

assign out_d = {sign_out_r2, exp_out[7:0], fract_out[23:1]};

always @(posedge clk)
	out <= #1 out_d;

always @(posedge clk)
	inf <= #1 inf_r;

always @(posedge clk)
	nan <= #1 nan_r;

endmodule































